# Copyright (c) QuantCo 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import pytest

import spox.opset.ai.onnx.v17 as op17
import spox.opset.ai.onnx.v19 as op19
import spox.opset.ai.onnx.v21 as op21
from spox import Optional, Sequence, Tensor, argument


@pytest.mark.parametrize("op", [op17, op19, op21])
def test_loop_inference(op):
    x, y, zs = op.loop(
        v_initial=[
            argument(Tensor(np.float64, (None,))),
            argument(Tensor(np.int64, ("N", 2))),
        ],
        body=lambda i, c, a, b: [op.const(True), a, op.add(i, b), i],
    )
    assert x.type == Tensor(np.float64, (None,))
    assert y.type == Tensor(np.int64, ("N", 2))
    assert zs.type == Tensor(np.int64, (None, 1))


@pytest.mark.parametrize("op", [op17, op19, op21])
def test_loop_concat(op):
    num_iters = op.const(1)
    v = op.const([], dtype=np.int64)

    result = op.loop(
        num_iters,
        v_initial=[v],
        body=lambda i, c, x: (op.const(True), op.concat([x, op.const([1])], axis=0)),
    )[0]

    # type can change, so we cannot infer anything
    assert result.type == Tensor(np.int64, None)


@pytest.mark.parametrize("op", [op17, op19, op21])
def test_loop_sequence(op):
    num_iters = op.const(1)
    v = op.sequence_empty(dtype=np.int64)

    result = op.loop(
        num_iters,
        v_initial=[v],
        body=lambda i, c, x: (op.const(True), op.sequence_insert(x, op.const([1]))),
    )[0]

    assert result.type == Sequence(Tensor(np.int64, None))


@pytest.mark.parametrize("op", [op17, op19, op21])
def test_loop_optional(op):
    num_iters = op.const(1)
    v = op.optional(type=Tensor(np.int64, (1, 2)))

    result = op.loop(
        num_iters,
        v_initial=[v],
        body=lambda i, c, x: (
            op.const(True),
            op.if_(
                op.optional_has_element(x),
                then_branch=lambda: [op.optional(type=Tensor(np.int64, (1, 2)))],
                else_branch=lambda: [op.optional(op.const([[1, 1]]))],
            )[0],
        ),
    )[0]

    assert result.type == Optional(Tensor(np.int64, (1, 2)))


@pytest.mark.parametrize("op", [op17, op19, op21])
def test_loop_optional_no_shape(op):
    num_iters = op.const(1)
    v = op.optional(type=Tensor(np.int64, (1, 2)))

    result = op.loop(
        num_iters,
        v_initial=[v],
        body=lambda i, c, x: (
            op.const(True),
            op.if_(
                op.optional_has_element(x),
                then_branch=lambda: [op.optional(type=Tensor(np.int64, (1, 2)))],
                else_branch=lambda: [op.optional(op.const([[1]]))],
            )[0],
        ),
    )[0]

    # shape can change, we cannot infer type
    assert result.type == Optional(Tensor(np.int64, None))



# Copyright (c) QuantCo 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Iterable
from dataclasses import dataclass

import numpy as np
import onnx
import onnx.parser
import onnxruntime as ort
import pytest
from onnx.numpy_helper import to_array

import spox.opset.ai.onnx.v18 as op18
import spox.opset.ai.onnx.v19 as op19
from spox import Tensor, Var, argument, build, inline
from spox._attributes import AttrInt64s
from spox._fields import BaseAttributes, BaseInputs, BaseOutputs
from spox._future import initializer
from spox._graph import arguments, results
from spox._node import OpType
from spox._standard import StandardNode
from spox._utils import make_model
from spox._var import _VarInfo


@pytest.fixture
def old_squeeze() -> onnx.ModelProto:
    return onnx.parser.parse_model(
        """
<
 ir_version: 8,
 opset_import: ["" : 12]
>
agraph (float[1, N] A) => (float[N] B)
{
    B = Squeeze<axes = [0]>(A)
}
"""
    )


@pytest.fixture
def old_identity() -> onnx.ModelProto:
    return onnx.parser.parse_model(
        """
<
 ir_version: 8,
 opset_import: ["" : 13]
>
agraph (float[N] A) => (float[N] B)
{
    B = Identity(A)
}
    """
    )


@pytest.fixture
def inline_old_squeeze_graph(old_squeeze):
    (data,) = arguments(
        data=Tensor(
            np.float32,
            (
                1,
                None,
            ),
        )
    )
    (result,) = inline(old_squeeze)(A=data).values()
    return results(final=result).with_opset(("ai.onnx", 17))


@pytest.fixture
def inline_old_identity_twice_graph(old_identity):
    (x,) = arguments(data=Tensor(np.float32, (None,)))
    (y,) = inline(old_identity)(A=x).values()
    (z,) = inline(old_identity)(A=y).values()
    return results(final=z).with_opset(("ai.onnx", 17))


class Squeeze11(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        axes: AttrInt64s

    @dataclass
    class Inputs(BaseInputs):
        data: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        squeezed: _VarInfo

    op_type = OpType("Squeeze", "", 11)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


def squeeze11(_data: Var, _axes: Iterable[int]):
    return (
        Squeeze11(
            Squeeze11.Attributes(AttrInt64s(_axes, "axes")),
            Squeeze11.Inputs(_data._var_info),
        )
        .get_output_vars()
        .squeezed
    )


@pytest.fixture
def old_squeeze_graph(old_squeeze):
    (data,) = arguments(
        data=Tensor(
            np.float32,
            (
                1,
                None,
            ),
        )
    )
    result = squeeze11(data, [0])
    return results(final=result).with_opset(("ai.onnx", 17))


def test_adapts_inline_old_squeeze(onnx_helper, inline_old_squeeze_graph):
    onnx_helper.assert_close(
        onnx_helper.run(
            inline_old_squeeze_graph,
            "final",
            data=np.array([[1, 2, 3, 4]], dtype=np.float32),
        ),
        [1, 2, 3, 4],
    )


def test_adapts_inline_old_identity_twice(onnx_helper, inline_old_identity_twice_graph):
    onnx_helper.assert_close(
        onnx_helper.run(
            inline_old_identity_twice_graph,
            "final",
            data=np.array([1, 2, 3, 4], dtype=np.float32),
        ),
        [1, 2, 3, 4],
    )


def test_adapts_singleton_old_squeeze(onnx_helper, old_squeeze_graph):
    onnx_helper.assert_close(
        onnx_helper.run(
            old_squeeze_graph,
            "final",
            data=np.array([[1, 2, 3, 4]], dtype=np.float32),
        ),
        [1, 2, 3, 4],
    )


def test_adapt_node_with_repeating_input_names():
    a = argument(Tensor(np.float32, ("N",)))
    b = op18.equal(a, a)
    c = op19.identity(a)

    build({"a": a}, {"b": b, "c": c})


def test_adapt_node_initializer():
    init_data = [1.0, 2.0, 3.0]

    a = argument(Tensor(np.float32, ("N",)))
    b = initializer(init_data, np.float32)
    c = op18.equal(a, b)
    d = op19.identity(a)

    model = build({"a": a}, {"b": b, "c": c, "d": d})
    np.testing.assert_allclose(to_array(model.graph.initializer[0]), init_data)


def test_inline_model_custom_node_only():
    """Inline a model which only consists of a custom node.

    Such models do not import from the default domain.
    """
    domain = "foo.ai"
    node = onnx.helper.make_node("FooOp", ["a"], ["b"], domain=domain)
    value_infos_input = [
        onnx.helper.make_value_info(
            "a", onnx.helper.make_tensor_type_proto(onnx.TensorProto.STRING, ("N",))
        ),
    ]
    value_infos_output = [
        onnx.helper.make_value_info(
            "b", onnx.helper.make_tensor_type_proto(onnx.TensorProto.STRING, ("N",))
        )
    ]

    model = make_model(
        onnx.helper.make_graph(
            [node],
            "graph",
            value_infos_input,
            value_infos_output,
        ),
        opset_imports=[onnx.helper.make_opsetid(domain, 1)],
    )

    # Ensure that our model is valid
    onnx.checker.check_model(model, full_check=True)

    (a,) = arguments(data=Tensor(np.str_, ("N",)))
    (b,) = inline(model)(a).values()

    # Add another node to the model to trigger the adaption logic
    c = op18.identity(b)
    build({"a": a}, {"c": c})


@pytest.mark.skip(
    reason="Adapting custom nodes (including their subgraphs) is currently not supported"
)
def test_inline_model_custom_node_nested(old_squeeze: onnx.ModelProto):
    """A singleton custom node with a old standard node in its attribute."""
    domain = "foo.ai"

    node = onnx.helper.make_node(
        "FooOp", ["a"], ["b"], domain=domain, **{"nested_graph": old_squeeze.graph}
    )
    value_infos_input = [
        onnx.helper.make_value_info(
            "a", onnx.helper.make_tensor_type_proto(onnx.TensorProto.FLOAT, ("N",))
        ),
    ]
    value_infos_output = [
        onnx.helper.make_value_info(
            "b", onnx.helper.make_tensor_type_proto(onnx.TensorProto.FLOAT, ("N",))
        )
    ]

    model = make_model(
        onnx.helper.make_graph(
            [node],
            "graph",
            value_infos_input,
            value_infos_output,
        ),
        opset_imports=[
            onnx.helper.make_opsetid(domain, 1),
            onnx.helper.make_opsetid("", 12),
        ],
    )

    # Ensure that our model is valid
    onnx.checker.check_model(model, full_check=True)

    (a,) = arguments(data=Tensor(np.float32, ("N",)))
    (b,) = inline(model)(a).values()

    # Add another node to the model to trigger the adaption logic
    c = op18.identity(b)
    build({"a": a}, {"c": c})


def test_if_adapatation_squeeze():
    cond = argument(Tensor(np.bool_, ()))
    b = argument(Tensor(np.float32, (1,)))
    squeezed = squeeze11(b, [0])
    out = op18.if_(
        cond,
        then_branch=lambda: [squeezed],
        else_branch=lambda: [squeeze11(b, [0])],
    )
    model = build({"b": b, "cond": cond}, {"out": out[0]})

    # predict on model
    b = np.array([1.1], dtype=np.float32)
    cond = np.array(True, dtype=np.bool_)
    out = ort.InferenceSession(model.SerializeToString()).run(
        None, {"b": b, "cond": cond}
    )


def test_if_adaptation_const():
    sq = op19.const(1.1453, dtype=np.float32)
    b = argument(Tensor(np.float32, ("N",)))
    cond = op18.equal(sq, b)
    out = op18.if_(cond, then_branch=lambda: [sq], else_branch=lambda: [sq])
    model = build({"b": b}, {"out": out[0]})
    assert model.domain == "" or model.domain == "ai.onnx"
    assert (
        model.opset_import[0].domain == "ai.onnx" or model.opset_import[0].domain == ""
    )
    assert model.opset_import[0].version > 11



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import pytest

import spox.opset.ai.onnx.ml.v3 as op_ml
from spox._graph import arguments
from spox._standard import InferenceError
from spox._type_system import Tensor


def test_scaler_inference():
    (x,) = arguments(x=Tensor(np.float64, ("N", 5)))
    y = op_ml.scaler(x, offset=[0.1, 0.2, 0.3, 0.4, 0.5], scale=[2.5])
    assert y.type == Tensor(np.float32, ("N", 5))


def test_scaler_inference_fails_mismatched_lengths():
    (x,) = arguments(x=Tensor(np.float64, ("N", 3)))
    with pytest.raises(InferenceError):
        op_ml.scaler(x, offset=[0.0, 0.1], scale=[1.0])



# Copyright (c) QuantCo 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Any

import numpy as np
import numpy.typing as npt
import onnx


def tensor_type_to_dtype(ttype: int) -> np.dtype:
    """Convert integer tensor types to ``numpy.dtype`` objects."""
    if ttype == onnx.TensorProto.STRING:
        return np.dtype(str)  # Spox uses the str datatype for strings, not object
    return onnx.helper.tensor_dtype_to_np_dtype(ttype)


def dtype_to_tensor_type(dtype_like: npt.DTypeLike) -> int:
    """Convert numpy data types into integer tensor types.

    Raises
    ------
    TypeError:
        If ``dtype_like`` has no corresponding tensor type in the ONNX standard.
    """
    err_msg = f"{dtype_like} is not a valid ONNX tensor element type."
    if dtype_like is None:  # numpy would default to float64
        raise TypeError(err_msg)
    # normalize in the case of aliases like ``long`` which are missing in the lookup
    dtype = np.dtype(np.dtype(dtype_like).type)
    if dtype == np.dtype(object):
        raise TypeError(
            "`np.dtype('object')` is not supported as a tensor element type. "
            "Hint: Spox uses `np.dtype('str')` for the string datatype."
        )
    elif dtype == np.dtype(str):
        return onnx.TensorProto.STRING
    try:
        return onnx.helper.np_dtype_to_tensor_dtype(dtype)
    except KeyError:
        raise TypeError(err_msg)


def make_model(
    graph: onnx.GraphProto,
    *,
    opset_imports: list[onnx.OperatorSetIdProto] | None = None,
    producer_name: str | None = None,
    doc_string: str | None = None,
    functions: list[onnx.FunctionProto] | None = None,
) -> onnx.ModelProto:
    """Like ``onnx.helper.make_model`` but with a consistent and conservative IR version that works on older runtimes."""
    ir_version = 8

    kwargs: dict[str, Any] = {}
    if opset_imports is not None:
        kwargs["opset_imports"] = opset_imports
    if producer_name is not None:
        kwargs["producer_name"] = producer_name
    if doc_string is not None:
        kwargs["doc_string"] = doc_string
    if functions is not None:
        kwargs["functions"] = functions

    return onnx.helper.make_model(graph, ir_version=ir_version, **kwargs)



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np

import spox.opset.ai.onnx.ml.v3 as op_ml
from spox._graph import arguments
from spox._type_system import Tensor


def test_binarizer_inference():
    (x,) = arguments(x=Tensor(np.float64, (None, 5, "N")))
    y = op_ml.binarizer(x)
    assert y.type == Tensor(np.float64, (None, 5, "N"))



# Copyright (c) QuantCo 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

"""Module containing experimental Spox features that may be standard in the future."""

from __future__ import annotations

import warnings
from collections.abc import Iterable, Iterator
from contextlib import contextmanager
from types import ModuleType
from typing import Any

import numpy as np
import numpy.typing as npt

import spox._node
import spox._value_prop
from spox._graph import initializer as _initializer
from spox._type_system import Tensor
from spox._var import Var

TypeWarningLevel = spox._node.TypeWarningLevel


def set_type_warning_level(level: TypeWarningLevel) -> None:
    spox._node._TYPE_WARNING_LEVEL = level


@contextmanager
def type_warning_level(level: TypeWarningLevel) -> Iterator[None]:
    prev_level = spox._node._TYPE_WARNING_LEVEL
    set_type_warning_level(level)
    yield
    set_type_warning_level(prev_level)


ValuePropBackend = spox._value_prop.ValuePropBackend


def set_value_prop_backend(backend: ValuePropBackend) -> None:
    spox._value_prop._VALUE_PROP_BACKEND = backend


@contextmanager
def value_prop_backend(backend: ValuePropBackend) -> Iterator[None]:
    prev_backend = spox._value_prop._VALUE_PROP_BACKEND
    set_value_prop_backend(backend)
    yield
    set_value_prop_backend(prev_backend)


def initializer(value: npt.ArrayLike, dtype: npt.DTypeLike | None = None) -> Var:
    """
    Create a Var with a constant value.

    Parameters
    ----------
    value
        Array-like value for the variable.
    dtype
        Data type for the given value. If ``None``, it is inferred from the value
        using numpy rules (``numpy.array(value)``).

    Returns
    -------
    Var
        Variable with the given constant ``value``.

    Notes
    -----
    When the model is built, constants created by this function become initializers.
    As such, they are independent of an opset version and are listed separately
    in the model. Initializers are also used internally in Spox.
    """
    return _initializer(np.array(value, dtype))


class _NumpyLikeOperatorDispatcher:
    def __init__(
        self, op: ModuleType, type_promotion: bool, constant_promotion: bool
    ) -> None:
        self.op = op
        self.type_promotion = type_promotion
        self.constant_promotion = constant_promotion

    def _promote(
        self, *args: Var | np.generic | int | float, to_floating: bool = False
    ) -> Iterable[Var | None]:
        """
        Apply constant promotion and type promotion to given parameters,
        creating constants and/or casting.
        """
        targets: list[np.dtype | np.generic | int | float] = [
            x.type.dtype if isinstance(x, Var) and isinstance(x.type, Tensor) else x  # type: ignore
            for x in args
        ]
        if self.type_promotion:
            target_type = np.dtype(np.result_type(*targets))
            if to_floating and not issubclass(target_type.type, np.floating):
                target_type = np.dtype("float64")
        else:
            dtypes = {dtype for dtype in targets if isinstance(dtype, np.dtype)}
            if len(dtypes) > 1:
                raise TypeError(
                    f"Inconsistent types for Var operator with no type promotion: {dtypes}."
                )
            (target_type,) = dtypes
            if issubclass(target_type.type, np.integer):
                disagreeing = [
                    value
                    for value in targets
                    if issubclass(np.result_type(value).type, np.floating)
                ]
                if disagreeing:
                    raise TypeError(
                        f"Floating constant operands promoting to integer type {target_type}: {disagreeing}"
                    )
            # TODO: Handle more constant-target inconsistencies here?

        def _promote_target(
            obj: Var | np.generic | int | float,
        ) -> Var | None:
            if self.constant_promotion and isinstance(obj, (np.generic, int, float)):
                return self.op.const(np.array(obj, dtype=target_type))
            elif isinstance(obj, Var):
                return self.op.cast(obj, to=target_type) if self.type_promotion else obj
            raise TypeError(
                f"Bad value '{obj!r}' of type {type(obj).__name__!r} for operator overloading with Var. "
                f"({self.type_promotion=}, {self.constant_promotion=})"
            )

        return tuple(var for var in map(_promote_target, args))

    def add(self, a, b) -> Var:  # type: ignore
        a, b = self._promote(a, b)
        return self.op.add(a, b)

    def sub(self, a, b) -> Var:  # type: ignore
        a, b = self._promote(a, b)
        return self.op.sub(a, b)

    def mul(self, a, b) -> Var:  # type: ignore
        a, b = self._promote(a, b)
        return self.op.mul(a, b)

    def truediv(self, a, b) -> Var:  # type: ignore
        a, b = self._promote(a, b, to_floating=True)
        return self.op.div(a, b)

    def floordiv(self, a, b) -> Var:  # type: ignore
        a, b = self._promote(a, b)
        c = self.op.div(a, b)
        if isinstance(c.type, Tensor) and not issubclass(c.type._elem_type, np.integer):
            c = self.op.floor(c)
        return c

    def neg(self, a: Var) -> Var:
        return self.op.neg(a)

    def and_(self, a: Var, b: Var) -> Var:
        return self.op.and_(a, b)

    def or_(self, a: Var, b: Var) -> Var:
        return self.op.or_(a, b)

    def xor(self, a: Var, b: Var) -> Var:
        return self.op.xor(a, b)

    def not_(self, a: Var) -> Var:
        return self.op.not_(a)


@contextmanager
def _operator_overloading(
    op: ModuleType, type_promotion: bool = False, constant_promotion: bool = True
) -> Iterator[None]:
    """Enable operator overloading on Var for this block.

    May be used either as a context manager, or a decorator.

    Semantics of operator overloading attempt to follow usual rules of Python and numpy.
    By default, type promotion is not enabled.

    Parameters
    ----------
    op
        An opset module of domain ``ai.onnx`` that will be used to construct the operator dispatcher.
        It's recommended to use a modern version, like 17, and rely on version converters.
    type_promotion
        Whether operator overloading should implicitly promote types according to numpy rules.
        Modifies the behaviour of true division to cast to ``float64`` before the operation
        if the type was not conclusively floating (as in numpy).
        False by default.
    constant_promotion
        Whether operator overloading should implicitly promote primitive scalar constants to Var.
        True by default.

    Examples
    --------
    >>> import numpy as np
    >>> from spox.opset.ai.onnx import v17 as op
    >>> x, y = op.const(2), op.const(3)
    >>> with operator_overloading(op):
    ...     z = x + y
    >>> assert z._get_value() == np.array(5)
    >>> @operator_overloading(op)
    ... def foo():
    ...    return x * y
    >>> assert foo()._get_value() == np.array(6)
    """
    prev_dispatcher = Var._operator_dispatcher
    Var._operator_dispatcher = _NumpyLikeOperatorDispatcher(
        op, type_promotion, constant_promotion
    )
    yield
    Var._operator_dispatcher = prev_dispatcher


def __getattr__(name: str) -> Any:
    if name == "operator_overloading":
        warnings.warn(
            "using 'operator_overloading' is deprecated, consider using https://github.com/Quantco/ndonnx instead",
            DeprecationWarning,
        )
        return _operator_overloading

    raise AttributeError(f"module {__name__} has no attribute {name}")


__all__ = [
    # Type warning levels
    "TypeWarningLevel",
    "set_type_warning_level",
    "type_warning_level",
    # Initializer-backed constants
    "initializer",
    # Value propagation backend
    "ValuePropBackend",
    "set_value_prop_backend",
    "value_prop_backend",
]



# Copyright (c) QuantCo 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

from spox._standard import InferenceError
from spox._type_system import Optional, Sequence, Tensor, Type


def loop_erase_shape_info(typ: Type) -> Type:
    """Erases the shape information for a type, that can exists as a state variable in a Loop"""
    if isinstance(typ, Tensor):
        return Tensor(typ.dtype, None)
    elif isinstance(typ, Sequence):
        if not isinstance(typ.elem_type, Tensor):
            raise InferenceError(
                f"Type {typ} not allowed for state variables in Loop operator, sequence element can only be a tensor"
            )
        return Sequence(loop_erase_shape_info(typ.elem_type))
    elif isinstance(typ, Optional):
        if isinstance(typ.elem_type, Optional):
            raise InferenceError(
                f"Type {typ} not allowed for state variables in Loop operator, optionals of optionals are not allowed"
            )
        return Optional(loop_erase_shape_info(typ.elem_type))
    raise InferenceError(
        f"Type {typ} not allowed for state variables in Loop operator."
    )



# Copyright (c) QuantCo 2023-2026
# SPDX-License-Identifier: BSD-3-Clause

# ruff: noqa: E741 -- Allow ambiguous variable name
from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass

import numpy as np

from spox._attributes import (
    AttrInt64,
    AttrInt64s,
    AttrTensor,
)
from spox._fields import BaseAttributes, BaseInputs, BaseOutputs
from spox._node import OpType
from spox._standard import StandardNode
from spox._var import (
    Var,
    _VarInfo,
    create_prop_dict,
    unwrap_vars,
)
from spox.opset.ai.onnx.ml.v4 import (
    _ArrayFeatureExtractor,
    _Binarizer,
    _CastMap,
    _CategoryMapper,
    _DictVectorizer,
    _FeatureVectorizer,
    _Imputer,
    _LabelEncoder,
    _LinearClassifier,
    _LinearRegressor,
    _Normalizer,
    _OneHotEncoder,
    _Scaler,
    _SVMClassifier,
    _SVMRegressor,
    _ZipMap,
    array_feature_extractor,
    binarizer,
    cast_map,
    category_mapper,
    dict_vectorizer,
    feature_vectorizer,
    imputer,
    label_encoder,
    linear_classifier,
    linear_regressor,
    normalizer,
    one_hot_encoder,
    scaler,
    svmclassifier,
    svmregressor,
    zip_map,
)


class _TreeEnsemble(StandardNode):
    @dataclass
    class Attributes(BaseAttributes):
        aggregate_function: AttrInt64
        leaf_targetids: AttrInt64s
        leaf_weights: AttrTensor
        membership_values: AttrTensor | None
        n_targets: AttrInt64 | None
        nodes_falseleafs: AttrInt64s
        nodes_falsenodeids: AttrInt64s
        nodes_featureids: AttrInt64s
        nodes_hitrates: AttrTensor | None
        nodes_missing_value_tracks_true: AttrInt64s | None
        nodes_modes: AttrTensor
        nodes_splits: AttrTensor
        nodes_trueleafs: AttrInt64s
        nodes_truenodeids: AttrInt64s
        post_transform: AttrInt64
        tree_roots: AttrInt64s

    @dataclass
    class Inputs(BaseInputs):
        X: _VarInfo

    @dataclass
    class Outputs(BaseOutputs):
        Y: _VarInfo

    op_type = OpType("TreeEnsemble", "ai.onnx.ml", 5)

    attrs: Attributes
    inputs: Inputs
    outputs: Outputs


def tree_ensemble(
    X: Var,
    *,
    aggregate_function: int = 1,
    leaf_targetids: Iterable[int],
    leaf_weights: np.ndarray,
    membership_values: np.ndarray | None = None,
    n_targets: int | None = None,
    nodes_falseleafs: Iterable[int],
    nodes_falsenodeids: Iterable[int],
    nodes_featureids: Iterable[int],
    nodes_hitrates: np.ndarray | None = None,
    nodes_missing_value_tracks_true: Iterable[int] | None = None,
    nodes_modes: np.ndarray,
    nodes_splits: np.ndarray,
    nodes_trueleafs: Iterable[int],
    nodes_truenodeids: Iterable[int],
    post_transform: int = 0,
    tree_roots: Iterable[int],
) -> Var:
    r"""
    Tree Ensemble operator. Returns the regressed values for each input in a
    batch. Inputs have dimensions ``[N, F]`` where ``N`` is the input batch
    size and ``F`` is the number of input features. Outputs have dimensions
    ``[N, num_targets]`` where ``N`` is the batch size and ``num_targets``
    is the number of targets, which is a configurable attribute.

    ::

       The encoding of this attribute is split along interior nodes and the leaves of the trees. Notably, attributes with the prefix `nodes_*` are associated with interior nodes, and attributes with the prefix `leaf_*` are associated with leaves.
       The attributes `nodes_*` must all have the same length and encode a sequence of tuples, as defined by taking all the `nodes_*` fields at a given position.

       All fields prefixed with `leaf_*` represent tree leaves, and similarly define tuples of leaves and must have identical length.

       This operator can be used to implement both the previous `TreeEnsembleRegressor` and `TreeEnsembleClassifier` nodes.
       The `TreeEnsembleRegressor` node maps directly to this node and requires changing how the nodes are represented.
       The `TreeEnsembleClassifier` node can be implemented by adding a `ArgMax` node after this node to determine the top class.
       To encode class labels, a `LabelEncoder` or `GatherND` operator may be used.

    Parameters
    ==========
    X
        Type T.
        Input of shape [Batch Size, Number of Features]
    aggregate_function
        Attribute.
        Defines how to aggregate leaf values within a target. One of 'AVERAGE'
        (0) 'SUM' (1) 'MIN' (2) 'MAX (3) defaults to 'SUM' (1)
    leaf_targetids
        Attribute.
        The index of the target that this leaf contributes to (this must be in
        range ``[0, n_targets)``).
    leaf_weights
        Attribute.
        The weight for each leaf.
    membership_values
        Attribute.
        Members to test membership of for each set membership node. List all of
        the members to test again in the order that the 'BRANCH_MEMBER' mode
        appears in ``node_modes``, delimited by ``NaN``\ s. Will have the same
        number of sets of values as nodes with mode 'BRANCH_MEMBER'. This may be
        omitted if the node doesn't contain any 'BRANCH_MEMBER' nodes.
    n_targets
        Attribute.
        The total number of targets.
    nodes_falseleafs
        Attribute.
        1 if false branch is leaf for each node and 0 if an interior node. To
        represent a tree that is a leaf (only has one node), one can do so by
        having a single ``nodes_*`` entry with true and false branches
        referencing the same ``leaf_*`` entry
    nodes_falsenodeids
        Attribute.
        If ``nodes_falseleafs`` is false at an entry, this represents the
        position of the false branch node. This position can be used to index
        into a ``nodes_*`` entry. If ``nodes_falseleafs`` is false, it is an
        index into the leaf\_\* attributes.
    nodes_featureids
        Attribute.
        Feature id for each node.
    nodes_hitrates
        Attribute.
        Popularity of each node, used for performance and may be omitted.
    nodes_missing_value_tracks_true
        Attribute.
        For each node, define whether to follow the true branch (if attribute
        value is 1) or false branch (if attribute value is 0) in the presence of
        a NaN input feature. This attribute may be left undefined and the
        default value is false (0) for all nodes.
    nodes_modes
        Attribute.
        The comparison operation performed by the node. This is encoded as an
        enumeration of 0 ('BRANCH_LEQ'), 1 ('BRANCH_LT'), 2 ('BRANCH_GTE'), 3
        ('BRANCH_GT'), 4 ('BRANCH_EQ'), 5 ('BRANCH_NEQ'), and 6
        ('BRANCH_MEMBER'). Note this is a tensor of type uint8.
    nodes_splits
        Attribute.
        Thresholds to do the splitting on for each node with mode that is not
        'BRANCH_MEMBER'.
    nodes_trueleafs
        Attribute.
        1 if true branch is leaf for each node and 0 an interior node. To
        represent a tree that is a leaf (only has one node), one can do so by
        having a single ``nodes_*`` entry with true and false branches
        referencing the same ``leaf_*`` entry
    nodes_truenodeids
        Attribute.
        If ``nodes_trueleafs`` is false at an entry, this represents the
        position of the true branch node. This position can be used to index
        into a ``nodes_*`` entry. If ``nodes_trueleafs`` is false, it is an
        index into the leaf\_\* attributes.
    post_transform
        Attribute.
        Indicates the transform to apply to the score. One of 'NONE' (0),
        'SOFTMAX' (1), 'LOGISTIC' (2), 'SOFTMAX_ZERO' (3) or 'PROBIT' (4),
        defaults to 'NONE' (0)
    tree_roots
        Attribute.
        Index into ``nodes_*`` for the root of each tree. The tree structure is
        derived from the branching of each node.

    Returns
    =======
    Y : Var
        Type T.
        Output of shape [Batch Size, Number of targets]

    Notes
    =====
    Signature: ``ai.onnx.ml@5::TreeEnsemble``.

    Type constraints:
     - T: `tensor(double)`, `tensor(float)`, `tensor(float16)`
    """
    input_prop_values = create_prop_dict(
        X=X,
    )
    output_vars = (
        _TreeEnsemble(
            _TreeEnsemble.Attributes(
                aggregate_function=AttrInt64(
                    aggregate_function, name="aggregate_function"
                ),
                leaf_targetids=AttrInt64s(leaf_targetids, name="leaf_targetids"),
                leaf_weights=AttrTensor(leaf_weights, name="leaf_weights"),
                membership_values=AttrTensor.maybe(
                    membership_values, name="membership_values"
                ),
                n_targets=AttrInt64.maybe(n_targets, name="n_targets"),
                nodes_falseleafs=AttrInt64s(nodes_falseleafs, name="nodes_falseleafs"),
                nodes_falsenodeids=AttrInt64s(
                    nodes_falsenodeids, name="nodes_falsenodeids"
                ),
                nodes_featureids=AttrInt64s(nodes_featureids, name="nodes_featureids"),
                nodes_hitrates=AttrTensor.maybe(nodes_hitrates, name="nodes_hitrates"),
                nodes_missing_value_tracks_true=AttrInt64s.maybe(
                    nodes_missing_value_tracks_true,
                    name="nodes_missing_value_tracks_true",
                ),
                nodes_modes=AttrTensor(nodes_modes, name="nodes_modes"),
                nodes_splits=AttrTensor(nodes_splits, name="nodes_splits"),
                nodes_trueleafs=AttrInt64s(nodes_trueleafs, name="nodes_trueleafs"),
                nodes_truenodeids=AttrInt64s(
                    nodes_truenodeids, name="nodes_truenodeids"
                ),
                post_transform=AttrInt64(post_transform, name="post_transform"),
                tree_roots=AttrInt64s(tree_roots, name="tree_roots"),
            ),
            _TreeEnsemble.Inputs(
                X=unwrap_vars(X),
            ),
        )
        .get_output_vars(input_prop_values=input_prop_values)
        .Y
    )
    return output_vars  # type: ignore


_OPERATORS = {
    "ArrayFeatureExtractor": _ArrayFeatureExtractor,
    "Binarizer": _Binarizer,
    "CastMap": _CastMap,
    "CategoryMapper": _CategoryMapper,
    "DictVectorizer": _DictVectorizer,
    "FeatureVectorizer": _FeatureVectorizer,
    "Imputer": _Imputer,
    "LabelEncoder": _LabelEncoder,
    "LinearClassifier": _LinearClassifier,
    "LinearRegressor": _LinearRegressor,
    "Normalizer": _Normalizer,
    "OneHotEncoder": _OneHotEncoder,
    "SVMClassifier": _SVMClassifier,
    "SVMRegressor": _SVMRegressor,
    "Scaler": _Scaler,
    "TreeEnsemble": _TreeEnsemble,
    "ZipMap": _ZipMap,
}

_CONSTRUCTORS = {
    "ArrayFeatureExtractor": array_feature_extractor,
    "Binarizer": binarizer,
    "CastMap": cast_map,
    "CategoryMapper": category_mapper,
    "DictVectorizer": dict_vectorizer,
    "FeatureVectorizer": feature_vectorizer,
    "Imputer": imputer,
    "LabelEncoder": label_encoder,
    "LinearClassifier": linear_classifier,
    "LinearRegressor": linear_regressor,
    "Normalizer": normalizer,
    "OneHotEncoder": one_hot_encoder,
    "SVMClassifier": svmclassifier,
    "SVMRegressor": svmregressor,
    "Scaler": scaler,
    "TreeEnsemble": tree_ensemble,
    "ZipMap": zip_map,
}

__all__ = [fun.__name__ for fun in _CONSTRUCTORS.values()]
